@@ -1,25 +1,49 @@
 package Test::BrewBuild;
-use 5.006;
 use strict;
 use warnings;
 
 use File::Temp;
+use Logging::Simple;
+use Test::BrewBuild::Plugin;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
+
+my $log;
 
 sub new {
     my ($class, %args) = @_;
     my $self = bless {}, $class;
     %{ $self->{args} } = %args;
+
+    $log = $self->_create_log($args{debug});
+    $log->_7("in new(), constructing " . __PACKAGE__ . " object");
+
+    my $exec_plugin_name = $args{plugin} ? $args{plugin} : $ENV{TBB_PLUGIN};
+
+    if ($exec_plugin_name){
+        $log->_7("plugin param set to: $exec_plugin_name");
+    }
+
+    $exec_plugin_name = $self->plugin($exec_plugin_name);
+
+    my $exec_plugin_sub = $exec_plugin_name .'::brewbuild_exec';
+    $self->{exec_plugin} = \&$exec_plugin_sub;
+
+    $log->_7("using plugin $exec_plugin_name");
+
     return $self;
 }
 sub perls_available {
     my ($self, $brew_info) = @_;
 
+    $log->_7("in perls_available()");
+
     my @perls_available = $self->is_win
         ? $brew_info =~ /(\d\.\d+\.\d+_\d+)/g
         : $brew_info =~ /(perl-\d\.\d+\.\d+)/g;
 
+    $log->_7("perls available: " . join ', ', @perls_available);
+
     if ($self->is_win){
         for (@perls_available){
             s/perl-//;
@@ -30,6 +54,9 @@ sub perls_available {
 sub perls_installed {
     my ($self, $brew_info) = @_;
 
+    my $log = $log->child('perls_installed');
+    $log->_7("checking perls installed");
+
     return $self->is_win
         ? $brew_info =~ /(\d\.\d{2}\.\d(?:_\d{2}))(?!=_)\s+\[installed\]/ig
         : $brew_info =~ /i.*?(perl-\d\.\d+\.\d+)/g;
@@ -37,28 +64,34 @@ sub perls_installed {
 sub instance_remove {
     my ($self, @perls_installed) = @_;
 
+    my $log = $log->child('instance_remove');
+
     if ($self->{args}{debug}) {
-        print "$_\n" for @perls_installed;
-        print "\nremoving previous installs...\n";
+        $log->_5("perls installed: " . join ', ', @perls_installed);
+        $log->_5("removing previous installs...");
     }
 
     my $remove_cmd = $self->is_win
         ? 'berrybrew remove'
         : 'perlbrew uninstall';
 
+    $log->_7("using $remove_cmd remove command");
+
     for (@perls_installed){
         my $ver = $^V;
         $ver =~ s/v//;
 
         if ($_ =~ /$ver$/){
-            print "skipping version we're using, $_\n" if $self->{args}{debug};
+            $log->_5("skipping version we're using, $_");
             next;
         }
+
+        $log->_7("exec'ing $remove_cmd");
+
         `$remove_cmd $_`;
     }
 
-    print "\nremoval of existing perl installs complete...\n"
-      if $self->{args}{debug};
+    $log->_5("removal of existing perl installs complete...");
 }
 sub instance_install {
     my $self = shift;
@@ -66,10 +99,14 @@ sub instance_install {
     my $perls_available = shift;
     my $perls_installed = shift;
 
+    my $log = $log->child('instance_install');
+
     my $install_cmd = $self->is_win
         ? 'berrybrew install'
         : 'perlbrew install --notest -j 4';
 
+    $log->_7("install cmd set to $install_cmd");
+
     my @new_installs;
 
     if ($self->{args}{version}->[0]){
@@ -79,9 +116,7 @@ sub instance_install {
                 : "perl-$version";
 
             if (grep { $version eq $_ } @{ $perls_installed }){
-                if ($self->{args}{debug}){
-                    warn "$version is already installed... skipping\n";
-                }
+                $log->_5("$version is already installed... skipping");
                 next;
             }
             push @new_installs, $version;
@@ -89,6 +124,9 @@ sub instance_install {
     }
     else {
         if ($count){
+
+            $log->_7("looking to install $count perl instances");
+
             while ($count > 0){
                 my $candidate = $perls_available->[rand @{ $perls_available }];
                 if (grep { $_ eq $candidate } @{ $perls_installed }) {
@@ -103,31 +141,35 @@ sub instance_install {
         }
     }
 
-    if ($self->{args}{debug}){
-        print "preparing to install...\n" if @new_installs;
-        print "$_\n" for @new_installs;
-    }
-
     if (@new_installs){
+        $log->_5("preparing to install..." . join ', ', @new_installs);
+
         for my $ver (@new_installs){
-            print "installing $ver...\n" if $self->{args}{debug};
+            $log->_5("installing $ver...");
+            $log->_7("...using cmd $install_cmd");
             `$install_cmd $ver`;
         }
     }
     else {
-        print "\nusing existing versions only\n" if $self->{args}{debug};
+        $log->_5("using existing versions only");
     }
 }
 sub results {
     my $self = shift;
 
+    my $log = $log->child('results');
+
     local $SIG{__WARN__} = sub {};
 
+    $log->_7("warnings trapped locally");
+
     my $result = $self->exec;
 
     my @ver_results = $result =~ /[Pp]erl-\d\.\d+\.\d+.*?Result:\s+\w+\n/gs;
 
-    print "\n\n";
+    $log->_7("got " . scalar @ver_results . " results");
+
+    my @pass;
 
     for (@ver_results){
         my $ver;
@@ -138,14 +180,23 @@ sub results {
         my $res;
 
         if (/Result:\s+(PASS)/){
+            $log->_7("$ver passed...");
             $res = $1;
         }
         else {
+            $log->_7("$ver FAILED...");
             print $_;
             exit;
         }
-        print "$ver :: $res\n";
+
+        $log->_7("$ver :: $res");
+        push @pass, "$ver :: $res\n";
     }
+
+    print "\n";
+    print $_ for @pass;
+
+    $log->_7(__PACKAGE__ ." run finished");
 }
 sub run {
     my $self = shift;
@@ -153,6 +204,9 @@ sub run {
 
     $count = 0 if ! $count;
 
+    my $log = $log->child('run');
+    $log->_7("commencing run()");
+
     my $brew_info = $self->brew_info;
 
     my @perls_available = $self->perls_available($brew_info);
@@ -161,12 +215,9 @@ sub run {
 
     my @perls_installed = $self->perls_installed($brew_info);
 
-    if ($self->{args}{debug}){
-        print "$_ installed\n" for @perls_installed;
-        print "\n";
-    }
+    $log->_5("installed perls: " . join ', ', @perls_installed);
 
-    $self->instance_remove(@perls_installed) if $self->{args}{reload};
+    $self->instance_remove(@perls_installed) if $self->{args}{remove};
     if ($count) {
         $self->instance_install($count, \@perls_available, \@perls_installed);
     }
@@ -175,39 +226,87 @@ sub run {
     @perls_installed = $self->perls_installed($brew_info);
 
     if (! @perls_installed) {
-        print "no perls installed... exiting\n";
+        $log->_0("no perls installed... exiting");
+        print "no perls installed... exiting" if $log->level;
         exit;
     }
 
     $self->results();
 }
 sub is_win {
-    return $^O =~ /Win/ ? 1 : 0;
+    my $log = $log->child("is_win");
+    my $is_win = ($^O =~ /Win/) ? 1 : 0;
+    $log->_7("is Windows: $is_win");
+    return $is_win;
 }
 sub exec {
     my (@a, @b);
     my $self = shift;
 
+    my $log = $log->child('exec');
+
+    $log->_7("creating temp file");
+
     my $wfh = File::Temp->new(UNLINK => 1);
     my $fname = $wfh->filename;
 
-    while(<DATA>){
+    $log->_7("temp filename: $fname");
+    $log->_7("fetching instructions from the plugin");
+    $log->_7("instructions to be executed:");
+
+    my @exec_cmd = $self->{exec_plugin}->();
+
+    for (@exec_cmd){
+        $log->_7($_);
         print $wfh $_;
     }
     close $wfh;
 
+    $log->_7("temp file handle closed");
+
     my $brew = $self->is_win ? 'berrybrew' : 'perlbrew';
+
+    $log->_7("exec'ing: $brew exec perl $fname");
+
     return `$brew exec perl $fname`;
 }
 sub brew_info {
     my $self = shift;
 
+    my $log = $log->child('brew_info');
+
     my $brew_info = $self->is_win
         ? `berrybrew available`
         : `perlbrew available`;
 
+    $log->_7("brew info set to:\n$brew_info");
+
     return $brew_info;
 }
+sub _create_log {
+    my ($self, $level) = @_;
+
+    $self->{log} = Logging::Simple->new(
+        name  => 'Test::BrewBuild',
+        level => $level,
+    );
+
+    $self->{log}->_7("in _create_log()");
+
+    if (defined $level && $level < 6){
+        $self->{log}->display(0);
+        $self->{log}->custom_display("-");
+        $self->{log}->_7("setting log level to $level");
+        $self->{log}->_7("log object created with level $level");
+    }
+
+    return $self->{log};
+}
+sub log {
+    my $self = shift;
+    $self->{log}->_7(ref($self) ." class/obj retrieving a log object");
+    return $self->{log};
+}
 1;
 
 =head1 NAME
@@ -215,6 +314,9 @@ sub brew_info {
 Test::BrewBuild - Perl/Berry brew unit testing automation across installed perl
 versions (Windows and Unix).
 
+=for html
+<a href="http://travis-ci.org/stevieb9/p5-test-brewbuild"><img src="https://secure.travis-ci.org/stevieb9/p5-test-brewbuild.png"/>
+<a href='https://coveralls.io/github/stevieb9/p5-test-brewbuild?branch=master'><img src='https://coveralls.io/repos/stevieb9/p5-test-brewbuild/badge.svg?branch=master&service=github' alt='Coverage Status' /></a>
 
 
 =head1 DESCRIPTION
@@ -246,7 +348,7 @@ All unit tests are run against all installed instances.
 
     my %args = (
         debug   => 0,
-        reload  => 0,
+        remove  => 0,
         version => '',
         count   => 0,
     );
@@ -277,6 +379,14 @@ All unit tests are run against all installed instances.
 Returns a new C<Test::BrewBuild> object. See the documentation for the
 C<berrybrew> script to understand what the arguments are and do.
 
+=head2 plugin('Module::Name')
+
+Fetches and installs a custom plugin which contains the code that
+C<perlbrew/berrybrew exec> will execute. If not used or the module specified
+can't be located (or it contains errors), we fall back to the default bundled
+L<Test::BrewBuild::Plugin::DefaultExec> (which is the canonical example for
+writing new plugins).
+
 =head2 perls_available
 
 Returns an array containing all perls available, whether already installed or
@@ -320,11 +430,22 @@ Helper method, returns true if the current OS is Windows, false if not.
 Helper method, returns the appropriate *brew calls relative to the platform
 we're working on.
 
+=head2 log
+
+Developer method, returns an instance of the packages log object for creating
+child log objects.
 
 =head1 AUTHOR
 
 Steve Bertrand, C<< <steveb at cpan.org> >>
 
+=head1 CONTRIBUTING
+
+Any and all feedback and help is appreciated. A Pull Request is the preferred
+method of receiving changes (L<https://github.com/stevieb9/p5-test-brewbuild>),
+but regular patches through the bug tracker, or even just email discussions are
+welcomed.
+
 =head1 BUGS
 
 L<https://github.com/stevieb9/p5-test-brewbuild/issues>
@@ -361,16 +482,4 @@ See L<http://dev.perl.org/licenses/> for more information.
 
 =cut
 
-1; # End of Test::BrewBuild
-
-__DATA__
-#!/usr/bin/perl
-use warnings;
-use strict;
-
-if ($^O ne 'MSWin32'){
-    system "cpanm --installdeps . && make && make test";
-}
-else {
-    system "cpanm --installdeps . && dmake && dmake test";
-}
+1;
